"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const iterall_1 = require("iterall");
const error_1 = require("graphql/error");
const invariant_1 = require("graphql/jsutils/invariant");
const isInvalid_1 = require("graphql/jsutils/isInvalid");
const isNullish_1 = require("graphql/jsutils/isNullish");
const memoize3_1 = require("graphql/jsutils/memoize3");
const promiseForObject_1 = require("graphql/jsutils/promiseForObject");
const promiseReduce_1 = require("graphql/jsutils/promiseReduce");
const values_1 = require("graphql/execution/values");
const definition_1 = require("graphql/type/definition");
const execute_1 = require("graphql/execution/execute");
const values_2 = require("graphql/execution/values");
const GraphQLDeferDirective_1 = require("./GraphQLDeferDirective");
const graphql_1 = require("graphql");
function isPromise(maybePromise) {
    return maybePromise && typeof maybePromise.then === 'function';
}
function buildExecutionContext(schema, document, rootValue, contextValue, rawVariableValues, operationName, fieldResolver, enableDefer) {
    const errors = [];
    let operation;
    let hasMultipleAssumedOperations = false;
    const fragments = Object.create(null);
    for (let i = 0; i < document.definitions.length; i++) {
        const definition = document.definitions[i];
        switch (definition.kind) {
            case graphql_1.Kind.OPERATION_DEFINITION:
                if (!operationName && operation) {
                    hasMultipleAssumedOperations = true;
                }
                else if (!operationName ||
                    (definition.name && definition.name.value === operationName)) {
                    operation = definition;
                }
                break;
            case graphql_1.Kind.FRAGMENT_DEFINITION:
                fragments[definition.name.value] = definition;
                break;
        }
    }
    if (!operation) {
        if (operationName) {
            errors.push(new error_1.GraphQLError(`Unknown operation named "${operationName}".`));
        }
        else {
            errors.push(new error_1.GraphQLError('Must provide an operation.'));
        }
    }
    else if (hasMultipleAssumedOperations) {
        errors.push(new error_1.GraphQLError('Must provide operation name if query contains ' +
            'multiple operations.'));
    }
    let variableValues;
    if (operation) {
        const coercedVariableValues = values_2.getVariableValues(schema, operation.variableDefinitions || [], rawVariableValues || {});
        if (coercedVariableValues.errors) {
            errors.push(...coercedVariableValues.errors);
        }
        else {
            variableValues = coercedVariableValues.coerced;
        }
    }
    if (errors.length !== 0) {
        return errors;
    }
    invariant_1.default(operation, 'Has operation if no errors.');
    invariant_1.default(variableValues, 'Has variables if no errors.');
    return {
        schema,
        fragments,
        rootValue,
        contextValue,
        operation: operation,
        variableValues,
        fieldResolver: fieldResolver || execute_1.defaultFieldResolver,
        errors,
        enableDefer,
    };
}
exports.buildExecutionContext = buildExecutionContext;
function shouldDeferNode(exeContext, node) {
    if (!exeContext.enableDefer) {
        return false;
    }
    const defer = values_1.getDirectiveValues(GraphQLDeferDirective_1.default, node, exeContext.variableValues);
    return defer !== undefined ? !defer.if : false;
}
function isDeferredExecutionResult(result) {
    return (result.initialResult !== undefined &&
        result.deferredPatches !== undefined);
}
exports.isDeferredExecutionResult = isDeferredExecutionResult;
function formatDataAsPatch(path, data, errors) {
    return {
        path: execute_1.responsePathAsArray(path),
        data,
        errors: errors && errors.length > 0 ? errors : undefined,
    };
}
function initializeDependentStore(exeContext, parentPath) {
    if (!exeContext.deferredDependents) {
        exeContext.deferredDependents = {};
    }
    if (!exeContext.deferredDependents[parentPath]) {
        exeContext.deferredDependents[parentPath] = {
            errors: [],
            patches: [],
        };
    }
}
function deferErrorToParent(exeContext, parentPath, error) {
    initializeDependentStore(exeContext, parentPath);
    exeContext.deferredDependents[parentPath].errors.push(error);
}
function deferPatchToParent(exeContext, parentPath, patch) {
    initializeDependentStore(exeContext, parentPath);
    exeContext.deferredDependents[parentPath].patches.push(patch);
}
function dispatchPatch(exeContext, patch) {
    if (!exeContext.patchDispatcher) {
        exeContext.patchDispatcher = new PatchDispatcher();
    }
    exeContext.patchDispatcher.dispatch(patch);
}
class PatchDispatcher {
    constructor() {
        this.resolvers = [];
        this.resultPromises = [];
    }
    dispatch(patch) {
        patch.then(({ patch, dependentPatches }) => {
            if (dependentPatches) {
                for (const patch of dependentPatches) {
                    this.dispatch(patch);
                }
            }
            const resolver = this.resolvers.shift();
            if (resolver)
                resolver({ value: patch, done: false });
        });
        this.resultPromises.push(new Promise(resolve => {
            this.resolvers.push(resolve);
        }));
    }
    getAsyncIterable() {
        const self = this;
        return {
            [iterall_1.$$asyncIterator]() {
                return {
                    next() {
                        return (self.resultPromises.shift() || Promise.resolve({ done: true }));
                    },
                };
            },
        };
    }
}
function execute(argsOrSchema, document, rootValue, contextValue, variableValues, operationName, fieldResolver, enableDefer) {
    return arguments.length === 1
        ? executeImpl(argsOrSchema.schema, argsOrSchema.document, argsOrSchema.rootValue, argsOrSchema.contextValue, argsOrSchema.variableValues, argsOrSchema.operationName, argsOrSchema.fieldResolver, argsOrSchema.enableDefer)
        : executeImpl(argsOrSchema, document, rootValue, contextValue, variableValues, operationName, fieldResolver, enableDefer);
}
exports.execute = execute;
function executeImpl(schema, document, rootValue, contextValue, variableValues, operationName, fieldResolver, enableDefer) {
    execute_1.assertValidExecutionArguments(schema, document, variableValues);
    const context = buildExecutionContext(schema, document, rootValue, contextValue, variableValues, operationName, fieldResolver, enableDefer);
    if (Array.isArray(context)) {
        return { errors: context };
    }
    const data = executeOperation(context, context.operation, rootValue);
    return buildResponse(context, data);
}
function buildResponse(context, data) {
    if (isPromise(data)) {
        return data.then(resolved => buildResponse(context, resolved));
    }
    const result = context.errors.length === 0 ? { data } : { errors: context.errors, data };
    if (context.patchDispatcher) {
        return {
            initialResult: result,
            deferredPatches: context.patchDispatcher.getAsyncIterable(),
        };
    }
    else {
        return result;
    }
}
function executeOperation(exeContext, operation, rootValue) {
    const type = getOperationRootType(exeContext.schema, operation);
    const fields = execute_1.collectFields(exeContext, type, operation.selectionSet, Object.create(null), Object.create(null));
    const path = undefined;
    try {
        const result = operation.operation === 'mutation'
            ? executeFieldsSerially(exeContext, type, rootValue, path, fields)
            : executeFields(exeContext, type, rootValue, path, fields);
        if (isPromise(result)) {
            return result.then(undefined, error => {
                exeContext.errors.push(error);
                return Promise.resolve(null);
            });
        }
        return result;
    }
    catch (error) {
        exeContext.errors.push(error);
        return null;
    }
}
function getOperationRootType(schema, operation) {
    switch (operation.operation) {
        case 'query':
            const queryType = schema.getQueryType();
            if (!queryType) {
                throw new error_1.GraphQLError('Schema does not define the required query root type.', [operation]);
            }
            return queryType;
        case 'mutation':
            const mutationType = schema.getMutationType();
            if (!mutationType) {
                throw new error_1.GraphQLError('Schema is not configured for mutations.', [
                    operation,
                ]);
            }
            return mutationType;
        case 'subscription':
            const subscriptionType = schema.getSubscriptionType();
            if (!subscriptionType) {
                throw new error_1.GraphQLError('Schema is not configured for subscriptions.', [
                    operation,
                ]);
            }
            return subscriptionType;
        default:
            throw new error_1.GraphQLError('Can only execute queries, mutations and subscriptions.', [operation]);
    }
}
exports.getOperationRootType = getOperationRootType;
function executeFieldsSerially(exeContext, parentType, sourceValue, path, fields) {
    return promiseReduce_1.default(Object.keys(fields), (results, responseName) => {
        const fieldNodes = fields[responseName];
        const fieldPath = execute_1.addPath(path, responseName);
        const result = resolveField(exeContext, parentType, sourceValue, fieldNodes, fieldPath);
        if (result === undefined) {
            return results;
        }
        if (isPromise(result)) {
            return result.then(resolvedResult => {
                results[responseName] = resolvedResult;
                return results;
            });
        }
        results[responseName] = result;
        return results;
    }, Object.create(null));
}
function executeFields(exeContext, parentType, sourceValue, path, fields, closestDeferredParent) {
    const results = Object.create(null);
    let containsPromise = false;
    for (let i = 0, keys = Object.keys(fields); i < keys.length; ++i) {
        const responseName = keys[i];
        const fieldNodes = fields[responseName];
        const fieldPath = execute_1.addPath(path, responseName);
        const result = resolveField(exeContext, parentType, sourceValue, fieldNodes, fieldPath, closestDeferredParent);
        if (result !== undefined) {
            results[responseName] = result;
            if (!containsPromise && isPromise(result)) {
                containsPromise = true;
            }
        }
    }
    if (!containsPromise) {
        return results;
    }
    return promiseForObject_1.default(results);
}
function resolveField(exeContext, parentType, source, fieldNodes, path, closestDeferredParent) {
    const fieldNode = fieldNodes[0];
    const fieldName = fieldNode.name.value;
    const fieldDef = execute_1.getFieldDef(exeContext.schema, parentType, fieldName);
    if (!fieldDef) {
        return;
    }
    const resolveFn = fieldDef.resolve || exeContext.fieldResolver;
    const info = execute_1.buildResolveInfo(exeContext, fieldDef, fieldNodes, parentType, path);
    const result = execute_1.resolveFieldValueOrError(exeContext, fieldDef, fieldNodes, resolveFn, source, info);
    return completeValueCatchingError(exeContext, fieldDef.type, fieldNodes, info, path, result, closestDeferredParent);
}
function asErrorInstance(error) {
    return error instanceof Error ? error : new Error(error || undefined);
}
function makePatchBundle(exeContext, path, data) {
    if (isPromise(data)) {
        return data.then(resolvedData => makePatchBundle(exeContext, path, resolvedData));
    }
    const dependent = exeContext.deferredDependents
        ? exeContext.deferredDependents[execute_1.responsePathAsArray(path).toString()]
        : undefined;
    return Promise.resolve({
        patch: formatDataAsPatch(path, data, dependent ? dependent.errors : []),
        dependentPatches: dependent ? dependent.patches : dependent,
    });
}
function completeValueCatchingError(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent) {
    const pathArray = execute_1.responsePathAsArray(path);
    const isListItem = typeof pathArray[pathArray.length - 1] === 'number';
    const shouldDefer = fieldNodes.every(node => shouldDeferNode(exeContext, node)) && !isListItem;
    if (definition_1.isNonNullType(returnType) && shouldDefer) {
        throw error_1.locatedError(new Error(`@defer cannot be applied on non-nullable field ${info.parentType.name}.${info.fieldName}`), fieldNodes, execute_1.responsePathAsArray(path));
    }
    const curClosestDeferredParent = shouldDefer
        ? execute_1.responsePathAsArray(path).toString()
        : closestDeferredParent;
    try {
        let completed;
        if (isPromise(result)) {
            completed = result.then(resolved => completeValue(exeContext, returnType, fieldNodes, info, path, resolved, curClosestDeferredParent));
        }
        else {
            completed = completeValue(exeContext, returnType, fieldNodes, info, path, result, curClosestDeferredParent);
        }
        if (shouldDefer) {
            let promisedPatch = makePatchBundle(exeContext, path, completed);
            if (closestDeferredParent) {
                deferPatchToParent(exeContext, closestDeferredParent, promisedPatch);
            }
            else {
                dispatchPatch(exeContext, promisedPatch);
            }
            return null;
        }
        if (isPromise(completed)) {
            return completed.then(undefined, error => {
                if (closestDeferredParent) {
                    handleDeferredFieldError(error, fieldNodes, path, returnType, exeContext, closestDeferredParent);
                    return null;
                }
                else {
                    return handleFieldError(error, fieldNodes, path, returnType, exeContext);
                }
            });
        }
        return completed;
    }
    catch (error) {
        if (closestDeferredParent || shouldDefer) {
            handleDeferredFieldError(error, fieldNodes, path, returnType, exeContext, closestDeferredParent);
            return null;
        }
        else {
            return handleFieldError(error, fieldNodes, path, returnType, exeContext);
        }
    }
}
function handleFieldError(rawError, fieldNodes, path, returnType, context) {
    const error = error_1.locatedError(asErrorInstance(rawError), fieldNodes, execute_1.responsePathAsArray(path));
    if (definition_1.isNonNullType(returnType)) {
        throw error;
    }
    context.errors.push(error);
    return null;
}
function handleDeferredFieldError(rawError, fieldNodes, path, _, exeContext, closestDeferredParent) {
    const error = error_1.locatedError(asErrorInstance(rawError), fieldNodes, execute_1.responsePathAsArray(path));
    const dependent = exeContext.deferredDependents
        ? exeContext.deferredDependents[execute_1.responsePathAsArray(path).toString()]
        : undefined;
    const shouldDefer = fieldNodes.every(node => shouldDeferNode(exeContext, node));
    if (shouldDefer) {
        const patch = formatDataAsPatch(path, null, [error]);
        const promisedPatch = Promise.resolve({
            patch,
            dependentPatches: dependent ? dependent.patches : dependent,
        });
        if (closestDeferredParent) {
            deferPatchToParent(exeContext, closestDeferredParent, promisedPatch);
        }
        else {
            dispatchPatch(exeContext, promisedPatch);
        }
    }
    if (closestDeferredParent) {
        deferErrorToParent(exeContext, closestDeferredParent, error);
    }
}
function completeValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent) {
    if (result instanceof Error) {
        throw result;
    }
    if (definition_1.isNonNullType(returnType)) {
        const completed = completeValue(exeContext, returnType.ofType, fieldNodes, info, path, result, closestDeferredParent);
        if (completed === null) {
            throw new Error(`Cannot return null for non-nullable field ${info.parentType.name}.${info.fieldName}.`);
        }
        return completed;
    }
    if (isNullish_1.default(result)) {
        return null;
    }
    if (definition_1.isListType(returnType)) {
        return completeListValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent);
    }
    if (definition_1.isLeafType(returnType)) {
        return completeLeafValue(returnType, result);
    }
    if (definition_1.isAbstractType(returnType)) {
        return completeAbstractValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent);
    }
    if (definition_1.isObjectType(returnType)) {
        return completeObjectValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent);
    }
    throw new Error(`Cannot complete value of unexpected type "${String(returnType)}".`);
}
function completeListValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent) {
    invariant_1.default(iterall_1.isCollection(result), `Expected Iterable, but did not find one for field ${info.parentType.name}.${info.fieldName}.`);
    const itemType = returnType.ofType;
    let containsPromise = false;
    const completedResults = [];
    iterall_1.forEach(result, (item, index) => {
        const fieldPath = execute_1.addPath(path, index);
        const completedItem = completeValueCatchingError(exeContext, itemType, fieldNodes, info, fieldPath, item, closestDeferredParent);
        if (!containsPromise && isPromise(completedItem)) {
            containsPromise = true;
        }
        completedResults.push(completedItem);
    });
    return containsPromise ? Promise.all(completedResults) : completedResults;
}
function completeLeafValue(returnType, result) {
    invariant_1.default(returnType.serialize, 'Missing serialize method on type');
    const serializedResult = returnType.serialize(result);
    if (isInvalid_1.default(serializedResult)) {
        throw new Error(`Expected a value of type "${String(returnType)}" but ` +
            `received: ${String(result)}`);
    }
    return serializedResult;
}
function completeAbstractValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent) {
    const runtimeType = returnType.resolveType
        ? returnType.resolveType(result, exeContext.contextValue, info)
        : defaultResolveTypeFn(result, exeContext.contextValue, info, returnType);
    if (isPromise(runtimeType)) {
        return runtimeType.then(resolvedRuntimeType => completeObjectValue(exeContext, ensureValidRuntimeType(resolvedRuntimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result, closestDeferredParent));
    }
    return completeObjectValue(exeContext, ensureValidRuntimeType(runtimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result, closestDeferredParent);
}
function ensureValidRuntimeType(runtimeTypeOrName, exeContext, returnType, fieldNodes, info, result) {
    const runtimeType = typeof runtimeTypeOrName === 'string'
        ? exeContext.schema.getType(runtimeTypeOrName)
        : runtimeTypeOrName;
    if (!definition_1.isObjectType(runtimeType)) {
        throw new error_1.GraphQLError(`Abstract type ${returnType.name} must resolve to an Object type at ` +
            `runtime for field ${info.parentType.name}.${info.fieldName} with ` +
            `value "${String(result)}", received "${String(runtimeType)}". ` +
            `Either the ${returnType.name} type should provide a "resolveType" ` +
            'function or each possible types should provide an ' +
            '"isTypeOf" function.', fieldNodes);
    }
    if (!exeContext.schema.isPossibleType(returnType, runtimeType)) {
        throw new error_1.GraphQLError(`Runtime Object type "${runtimeType.name}" is not a possible type ` + `for "${returnType.name}".`, fieldNodes);
    }
    return runtimeType;
}
function completeObjectValue(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent) {
    if (returnType.isTypeOf) {
        const isTypeOf = returnType.isTypeOf(result, exeContext.contextValue, info);
        if (isPromise(isTypeOf)) {
            return isTypeOf.then(resolvedIsTypeOf => {
                if (!resolvedIsTypeOf) {
                    throw invalidReturnTypeError(returnType, result, fieldNodes);
                }
                return collectAndExecuteSubfields(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent);
            });
        }
        if (!isTypeOf) {
            throw invalidReturnTypeError(returnType, result, fieldNodes);
        }
    }
    return collectAndExecuteSubfields(exeContext, returnType, fieldNodes, info, path, result, closestDeferredParent);
}
function invalidReturnTypeError(returnType, result, fieldNodes) {
    return new error_1.GraphQLError(`Expected value of type "${returnType.name}" but got: ${String(result)}.`, fieldNodes);
}
function collectAndExecuteSubfields(exeContext, returnType, fieldNodes, _, path, result, closestDeferredParent) {
    const subFieldNodes = collectSubfields(exeContext, returnType, fieldNodes);
    return executeFields(exeContext, returnType, result, path, subFieldNodes, closestDeferredParent);
}
const collectSubfields = memoize3_1.default(_collectSubfields);
function _collectSubfields(exeContext, returnType, fieldNodes) {
    let subFieldNodes = Object.create(null);
    const visitedFragmentNames = Object.create(null);
    for (let i = 0; i < fieldNodes.length; i++) {
        const selectionSet = fieldNodes[i].selectionSet;
        if (selectionSet) {
            subFieldNodes = execute_1.collectFields(exeContext, returnType, selectionSet, subFieldNodes, visitedFragmentNames);
        }
    }
    return subFieldNodes;
}
function defaultResolveTypeFn(value, context, info, abstractType) {
    if (value !== null &&
        typeof value === 'object' &&
        typeof value.__typename === 'string') {
        return value.__typename;
    }
    const possibleTypes = info.schema.getPossibleTypes(abstractType);
    const promisedIsTypeOfResults = [];
    for (let i = 0; i < possibleTypes.length; i++) {
        const type = possibleTypes[i];
        if (type.isTypeOf) {
            const isTypeOfResult = type.isTypeOf(value, context, info);
            if (isPromise(isTypeOfResult)) {
                promisedIsTypeOfResults[i] = isTypeOfResult;
            }
            else if (isTypeOfResult) {
                return type;
            }
        }
    }
    if (promisedIsTypeOfResults.length) {
        return Promise.all(promisedIsTypeOfResults).then(isTypeOfResults => {
            for (let i = 0; i < isTypeOfResults.length; i++) {
                if (isTypeOfResults[i]) {
                    return possibleTypes[i];
                }
            }
            return;
        });
    }
    return;
}
//# sourceMappingURL=execute.js.map